"""Ol Browser Login Utility"""
##################################################
# MIT License
##################################################
# File: olbrowserlogin.py
# Description: Overleaf Browser Login Utility
# Author: Moritz Glöckl
# License: MIT
# Version: 1.2.0
##################################################

from PySide6.QtCore import *
from PySide6.QtWidgets import *
from PySide6.QtWebEngineWidgets import *
from PySide6.QtWebEngineCore import QWebEngineProfile, QWebEngineSettings, QWebEnginePage


# # Where to get the CSRF Token and where to send the login request to
# LOGIN_URL = "https://www.overleaf.com/login"
# PROJECT_URL = "https://www.overleaf.com/project"  # The dashboard URL

# JS snippet to get the first link
JAVASCRIPT_EXTRACT_PROJECT_URL = "document.getElementsByClassName('dash-cell-name')[1].firstChild.href"
# JS snippet to extract the csrfToken
JAVASCRIPT_CSRF_EXTRACTOR = "document.getElementsByName('ol-csrfToken')[0].content"

class OlBrowserLoginWindow(QMainWindow):
    """
    Overleaf Browser Login Utility
    Opens a browser window to securely login the user and returns relevant login data.
    """

    def __init__(self, ce_url=None, *args, **kwargs):
        super(OlBrowserLoginWindow, self).__init__(*args, **kwargs)

        if ce_url is not None:
            self._BASE_URL = ce_url
            self._cookies_names = ["sharelatex.sid"]
        else:
            self._BASE_URL = "https://www.overleaf.com"  # The Overleaf Base URL
            self._cookies_names = ["overleaf_session2", "GCLB"]

        self._LOGIN_URL = f"{self._BASE_URL}/login"
        self._PROJECT_URL = f"{self._BASE_URL}/project"

        self.webview = QWebEngineView()

        self._cookies = {}
        self._csrf = ""
        self._login_success = False

        self.profile = QWebEngineProfile(self.webview)
        self.cookie_store = self.profile.cookieStore()
        self.cookie_store.cookieAdded.connect(self.handle_cookie_added)
        self.profile.setPersistentCookiesPolicy(QWebEngineProfile.NoPersistentCookies)

        self.profile.settings().setAttribute(QWebEngineSettings.JavascriptEnabled, True)

        webpage = QWebEnginePage(self.profile, self)
        self.webview.setPage(webpage)
        self.webview.load(QUrl.fromUserInput(self._LOGIN_URL))
        self.webview.loadFinished.connect(self.handle_load_finished)

        self.setCentralWidget(self.webview)
        self.resize(600, 700)

    def handle_load_finished(self):
         def callback(result):

             def callback(result):
                 self._csrf = result
                 self._login_success = True
                 QCoreApplication.quit()

             self.webview.load(QUrl.fromUserInput(result))
             self.webview.loadFinished.connect( lambda x:
                 self.webview.page().runJavaScript(
                     JAVASCRIPT_CSRF_EXTRACTOR, 0, callback
                 )
             )

         if self.webview.url().toString() == self._PROJECT_URL:
             self.webview.page().runJavaScript(
                 JAVASCRIPT_EXTRACT_PROJECT_URL, 0, callback
             )

    def handle_cookie_added(self, cookie):
        cookie_name = cookie.name().data().decode('utf-8')
        if cookie_name in self._cookies_names:
            self._cookies[cookie_name] = cookie.value().data().decode('utf-8')

    @property
    def cookies(self):
        return self._cookies

    @property
    def csrf(self):
        return self._csrf

    @property
    def login_success(self):
        return self._login_success


def login(ce_url=None):
    from PySide6.QtCore import QLoggingCategory
    QLoggingCategory.setFilterRules('''\
    qt.webenginecontext.info=false
    ''')

    app = QApplication([])
    ol_browser_login_window = OlBrowserLoginWindow(ce_url)
    ol_browser_login_window.show()
    app.exec()

    if not ol_browser_login_window.login_success:
        return None

    return {"cookie": ol_browser_login_window.cookies, "csrf": ol_browser_login_window.csrf}
